/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <boost/optional.hpp>
#include <boost/lexical_cast.hpp>

#include <tools/TaskScheduler.h>
#include <whiteWalls/FPNCalibration.h>
#include <boost/format.hpp>
#include <ImfThreading.h>


#include <string>
#include <thread>

int main(int argc, char **argv)
{
    Imf::setGlobalThreadCount(std::thread::hardware_concurrency());
    std::cout << "Running with " << std::thread::hardware_concurrency() << " threads" << std::endl;
    TaskScheduler::Init(std::thread::hardware_concurrency());

    /*
     * This vector's size is equal to the number of different shutter speeds for which images were captured.
     * For each shutter speed, the vector contains a vector of filenames, one for each image with that shutter speed.
     */
    std::vector<std::vector<std::string>> filenames;
#if 0
    filenames.resize(8);
    for (unsigned i = 0; i < 8; i++)
        for (unsigned j = 0; j < 88; j++)
            filenames[i].push_back((boost::format("/some/where/on/your/drive/DSC%05d.tiff") % (1303+i+j*8)).str());
#else
    std::cerr << "You must hack the selection of the calibration dataset into the main.cpp file." << std::endl;
    return -1;
#endif

    whiteWalls::FPNCalibration calibData(std::move(filenames));

    /*
     * The individual steps are individual functions that write their results to temporary files.
     * This allows to comment them out if experimentation with later stages is performed.
     */
    calibData.computeMeanImage();
    calibData.smoothAvgImage();

    calibData.computeVariance();
    //calibData.writeModelDebugPlots(std::cout);

    whiteWalls::Image3f scales, offsets;
    calibData.computeLinearModel(scales, offsets);

    /*
    In case you want to produce histograms to look at the range of offsets/scales. Plot with matlab/octave.
    std::cout << "figure(1);" << std::endl;
    whiteWalls::FPNCalibration::produceHistogram(offsets, -20.0f, 20.0f, 50, std::cout);
    std::cout << "figure(2);" << std::endl;
    whiteWalls::FPNCalibration::produceHistogram(scales, 0.99, 1.01, 50, std::cout);
    */


    /*
     * These are the files that can be passed to the denoising executable.
     */
    whiteWalls::writeExrImage("offset.exr", offsets, Imf::globalThreadCount());
    whiteWalls::writeExrImage("scale.exr", scales, Imf::globalThreadCount());

    /*
     * These are just for visualization and not actually needed.
     */
    whiteWalls::FPNCalibration::produceDebugImgOfOffset(offsets, "offset.png");
    whiteWalls::FPNCalibration::produceDebugImgOfScale(scales, "scale.png");



    return 0;
}
