/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <boost/optional.hpp>
#include <boost/lexical_cast.hpp>

#include <tools/RasterImage.h>
#include <tools/TaskScheduler.h>

#include <whiteWalls/CameraResponseCurve.h>
#include <whiteWalls/ImageHelpers.h>
#include <whiteWalls/Denoiser.h>
#include <ImfThreading.h>

#include <string>
#include <vector>
#include <Eigen/Dense>
#include <fstream>
#include <thread>

void showUsage()
{
    std::cout
         << "Usage: imgStackDenoiser {--output outputFilename.png | --output-pre-wallis preWallis.exr } --sigma 10.0 [--non-linear] [--FPNOffset fpnOffset.exr --FPNScale fpnScale.exr] [--fullCovar] --filenames files/*.tiff" << std::endl
        << "    --output : if/where (filename) to store the wallis filtered 8-bit image. Can also be a lossy .jpg!" << std::endl
        << "    --output-pre-wallis : if/where (filename) to store the unfiltered floating point colors (always stores as .exr file)" << std::endl
        << "        One of --output and --output-pre-wallis is mandatory" << std::endl
        << "    --sigma : size of the gaussian filter kernel (in pixels) for all smoothing operations" << std::endl
        << "    --non-linear : input images are not 16-bit raw conversions but tonemapped 8-bit images. Note that tonemapping curves from the Eos400D are assumed and no inverse color matrix is applied!" << std::endl
        << "    --FPNOffset and --FPNScale : load and use fixed pattern noise data" << std::endl
        << "    --fullCovar : don't normalize color channels independently. We didn't see any benefit from this." << std::endl
        << "    --filenames : list of input filenames" << std::endl
        << std::endl
        << "Typical use case:" << std::endl
        << "1. Use dcraw to convert and debayer raw files:" << std::endl
        << "   dcraw -t 0 -T -4 -q 3 -M -o 0 vantagePoint0/*.dng" << std::endl
        << "(-t 0: disable all flipping/rotation; -T: write tiff; -4: 16-bit output and no gamma; -q 3: high quality debayer; -M: Don't use color matrix; -o 0: keep original color space)" << std::endl
        << "2. Use this tool to stack and denoise:" << std::endl
        << "   imgStackDenoiser --output processed/img0.jpg --sigma 10 --filenames vantagePoint0/*.tiff" << std::endl
        << "or if you have a FPN model stored in the camera/ subfolder:" << std::endl
        << "   imgStackDenoiser --output processed/img0.jpg --sigma 10 --FPNOffset camera/fpnOffset.exr --FPNScale camera/fpnScale.exr --filenames vantagePoint0/*.tiff" << std::endl
        << "3. Repeat for vantagePoint1 .. vantagePointN" << std::endl;

}


int main(int argc, char **argv)
{
    Imf::setGlobalThreadCount(std::thread::hardware_concurrency());
    TaskScheduler::Init(std::thread::hardware_concurrency());


    std::cout << "ImageStackDenoiser tool" << std::endl;


    boost::optional<float> sigma = boost::make_optional<float>(false, 0.0f); // prevent gcc's uninitialized variable warning

    bool linear = true;
    bool fullCovar = false;
    std::vector<std::string> filenames;
    std::string FPNOffsetFilename;
    std::string FPNScaleFilename;
    boost::optional<std::string> outputFilename;
    boost::optional<std::string> outputPreWallisFilename;


    for (int i = 1; i < argc; i++) {
        std::string arg(argv[i]);

        if (arg == "--output") {
            if (i+1 == argc) {
                std::cerr << "Missing argument after --sigma" << std::endl;
                exit(-1);
            }
            i++;
            outputFilename = argv[i];
        } else
        if (arg == "--sigma") {
            if (i+1 == argc) {
                std::cerr << "Missing argument after --sigma" << std::endl;
                exit(-1);
            }
            try {
                sigma = boost::lexical_cast<float>(argv[i+1]);
            } catch (...) {
                std::cerr << "Could not cast '" << argv[i+1] << "' to float number for --sigma"  << std::endl;
                exit(-1);
            }
            i++;
        } else
        if (arg == "--non-linear") {
            linear = false;
        } else
        if (arg == "--output-pre-wallis") {
            if (i+1 == argc) {
                std::cerr << "Missing argument after --output-pre-wallis" << std::endl;
                exit(-1);
            }
            i++;
            outputPreWallisFilename = argv[i];
        } else
        if (arg == "--FPNOffset") {
            if (i+1 == argc) {
                std::cerr << "Missing argument after --sigma" << std::endl;
                exit(-1);
            }
            FPNOffsetFilename = argv[i+1];
            i++;
        } else
        if (arg == "--FPNScale") {
            if (i+1 == argc) {
                std::cerr << "Missing argument after --sigma" << std::endl;
                exit(-1);
            }
            FPNScaleFilename = argv[i+1];
            i++;
        } else
        if (arg == "--fullCovar") {
            fullCovar = true;
        } else
        if (arg == "--filenames") {
            if (i+1 == argc) {
                std::cerr << "Missing argument(s) after --filenames" << std::endl;
                exit(-1);
            }
            i++;
            for (; i < argc; i++)
                filenames.push_back(argv[i]);
        } else {
            std::cerr << "Unknown argument '" << argv[i] << "'" << std::endl;
            exit(-1);
        }
    }

    if (!linear) {
        std::cout << "Warning: Non-linear images are automatically converted into a linear space with arbitraty tonemapping curves from an Eos400D."
                  << "We did not do extensive tests whether linear colorspaces are actually better than sRGB color spaces because using RAW just worked so much better."
                  << "It might be beneficial to not do the conversion by declaring JPG images as 'linear' and dropping the --non-linear argument." << std::endl;
    }

    if (!outputFilename && !outputPreWallisFilename) {
        std::cerr << "output or output-pre-wallis is mandatory!" << std::endl;
        showUsage();
        exit(-1);
    }

    if (filenames.empty()) {
        std::cerr << "No input filenames specified!" << std::endl;
        showUsage();
        exit(-1);
    }

    if (!sigma) {
        std::cerr << "sigma is mandatory!" << std::endl;
        showUsage();
        exit(-1);
    }
    
    if (FPNOffsetFilename.empty() != FPNScaleFilename.empty()) {
        std::cerr << "Fixed pattern noise removal needs offset and scale exr files!" << std::endl;
        showUsage();
        exit(-1);
    }
    
    bool hasFPN = !FPNOffsetFilename.empty();

    whiteWalls::Image3f fpnOffset, fpnScale;

    if (hasFPN) {
        whiteWalls::readExrImage(FPNOffsetFilename, fpnOffset);
        whiteWalls::readExrImage(FPNScaleFilename, fpnScale);
    }

    whiteWalls::Denoiser denoiser(*sigma, std::move(fpnScale), std::move(fpnOffset));

    std::cout << "Computing denoised image: " << std::endl;
    denoiser.denoise(filenames, linear, [](const std::string &msg) {
        std::cout << msg << std::endl;
    });

    if (outputPreWallisFilename) {
        std::cout << "Writing pre-wallis exr file..." << std::endl;
        whiteWalls::writeExrImage(*outputPreWallisFilename, denoiser.getDenoisedImage());
    }

    if (outputFilename) {
        std::cout << "Running augmented wallis filter: " << std::endl;
        RasterImage wallisFilteredImage = denoiser.wallisFilter(filenames, linear, fullCovar, [](const std::string &msg) {
            std::cout << msg << std::endl;
        });

        std::cout << "Writing wallis filtered image..." << std::endl;
        wallisFilteredImage.writeToFile(outputFilename->c_str(), 100);
    }

    return 0;
}
