/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef CAMERARESPONSECURVE_H
#define CAMERARESPONSECURVE_H

#include <initializer_list>
#include <algorithm>
#include <assert.h>

namespace synthBench {
namespace math {

template<unsigned numLdrValues>
class CameraResponseCurve
{
    public:
        CameraResponseCurve() {
            for (unsigned i = 0; i < numLdrValues; i++)
                m_hdrValues[i] = i * 2.0f / (numLdrValues - 1);
        }

        CameraResponseCurve(std::initializer_list<float> values) {
            assert((values.size() == numLdrValues) && "Number of values in initializer list must match numLdrValues");
            for (unsigned i = 0; i < numLdrValues; i++)
                m_hdrValues[i] = values.begin()[i];
        }

        float &ldrToHdr(unsigned ldr) { return m_hdrValues[ldr]; }
        const float &ldrToHdr(unsigned ldr) const { return m_hdrValues[ldr]; }

        unsigned ldrFromHdr(float hdr) const;
    protected:
        float m_hdrValues[numLdrValues];
};


template<unsigned numLdrValues>
unsigned CameraResponseCurve<numLdrValues>::ldrFromHdr(float hdr) const
{
    int min = 0;
    int max = numLdrValues-1;
    while (min < max) {
        int center = (max + min)/2;
        float centerValue = m_hdrValues[center];
        if (hdr < centerValue)
            max = center-1;
        else
            min = center+1;
    }
    return std::max<int>(std::min<int>(min, numLdrValues-1), 0);
}


namespace Eos400D_neutral_5200K_daylight {
extern synthBench::math::CameraResponseCurve<256> redCurve;
extern synthBench::math::CameraResponseCurve<256> greenCurve;
extern synthBench::math::CameraResponseCurve<256> blueCurve;
}

namespace Eos400D_neutral_3200K_artificial {
extern synthBench::math::CameraResponseCurve<256> redCurve;
extern synthBench::math::CameraResponseCurve<256> greenCurve;
extern synthBench::math::CameraResponseCurve<256> blueCurve;
}

namespace Eos400D_neutral_7000K_shade {
extern synthBench::math::CameraResponseCurve<256> redCurve;
extern synthBench::math::CameraResponseCurve<256> greenCurve;
extern synthBench::math::CameraResponseCurve<256> blueCurve;
}


}
}


#endif // CAMERARESPONSECURVE_H
