/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "Denoiser.h"
#include <stdexcept>

namespace whiteWalls {

Denoiser::Denoiser(float sigma, Image3f fpnScale, Image3f fpnOffset) :
            m_fpnScale(std::move(fpnScale)),
            m_fpnOffset(std::move(fpnOffset))
{
    m_kernel = buildGaussianKernel(sigma);
    
    if ((fpnOffset.rows() != fpnScale.rows()) ||
        (fpnOffset.cols() != fpnScale.cols()))
        throw std::runtime_error("fpn image dimensions mismatch!");    
}

void Denoiser::denoise(const std::vector<std::string> &filenames, bool linear, ProgressCallback progressCallback)
{
    unsigned width, height;
    RasterImage::getImageFileDimensions(filenames[0].c_str(), width, height);
    
    bool hasFPN = m_fpnScale.rows() != 0;
    
    if (hasFPN) {
        if ((m_fpnScale.cols() != width) ||
            (m_fpnScale.rows() != height)) {
            throw std::runtime_error("FPN image sizes do not match the input images!");
        }
    }
    
    auto readImg = [linear, width, height](const std::string &imageFilename, Image3f &img) {
        if (linear)
            readLinear16BitImage(imageFilename, img);
        else
            readTonemappedImage(imageFilename, img,
                                synthBench::math::Eos400D_neutral_5200K_daylight::redCurve,
                                synthBench::math::Eos400D_neutral_5200K_daylight::greenCurve,
                                synthBench::math::Eos400D_neutral_5200K_daylight::blueCurve);
            
        if ((img.cols() != width) ||
            (img.rows() != height)) {
            throw std::runtime_error("All images must have the same size!");
        }
    };

    Image3f img;
    m_denoised.resize(height, width);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            m_denoised(y, x).setZero();


    for (const std::string &imageFilename : filenames) {
        if (progressCallback)
            progressCallback(std::string("Processing ") + imageFilename);
        readImg(imageFilename, img);
        if (hasFPN)
            removeLinearFPN(img, m_fpnScale, m_fpnOffset);

        m_denoised += img;
    }

    float rcpSampleCount = 1.0f / filenames.size();
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            m_denoised(y, x) *= rcpSampleCount;
}


RasterImage Denoiser::wallisFilter(const std::vector<std::string> &filenames, bool linear, bool fullCovar, ProgressCallback progressCallback)
{
    unsigned width = m_denoised.cols();
    unsigned height = m_denoised.rows();
    
    Image3f smoothAverageImg;
    smoothImage(smoothAverageImg, m_denoised, m_kernel);

    Image3f smoothImg, noiseVarImg;
    noiseVarImg.resize(height, width);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            noiseVarImg(y, x).setZero();

    auto readImg = [linear, width, height](const std::string &imageFilename, Image3f &img) {
        if (linear)
            readLinear16BitImage(imageFilename, img);
        else
            readTonemappedImage(imageFilename, img,
                                synthBench::math::Eos400D_neutral_5200K_daylight::redCurve,
                                synthBench::math::Eos400D_neutral_5200K_daylight::greenCurve,
                                synthBench::math::Eos400D_neutral_5200K_daylight::blueCurve);
            
        if ((img.cols() != width) ||
            (img.rows() != height)) {
            throw std::runtime_error("All images must have the same size!");
        }
    };

    Image3f img;
    for (const std::string &imageFilename : filenames) {
        if (progressCallback)
            progressCallback(std::string("Processing ") + imageFilename);
        readImg(imageFilename, img);

        smoothImage(smoothImg, img, m_kernel);

        for (unsigned y = 0; y < height; y++)
            for (unsigned x = 0; x < width; x++)
                for (unsigned c = 0; c < 3; c++) {
                    float adjusted = img(y, x)[c] * smoothAverageImg(y, x)[c] / std::max(1.0f, smoothImg(y, x)[c]);
                    noiseVarImg(y, x)[c] += (adjusted - m_denoised(y, x)[c])*(adjusted - m_denoised(y, x)[c]);
                }
    }

    float rcpSampleCount = 1.0f / filenames.size();
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            noiseVarImg(y, x) *= rcpSampleCount;

    m_remainingNoiseVariance.resize(height, width);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            m_remainingNoiseVariance(y, x) = noiseVarImg(y, x) * rcpSampleCount;


    Image3f sqrAverageImg = m_denoised*m_denoised;
    sqrAverageImg.resize(height, width);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++) {
            sqrAverageImg(y, x) = m_denoised(y, x).cwiseProduct(m_denoised(y, x));
        }

    Image3f covAverageImg;
    if (fullCovar) {
        covAverageImg.resize(height, width);
        for (unsigned y = 0; y < height; y++)
            for (unsigned x = 0; x < width; x++) {
                covAverageImg(y, x)[0] = m_denoised(y, x)[0] * m_denoised(y, x)[1];
                covAverageImg(y, x)[1] = m_denoised(y, x)[0] * m_denoised(y, x)[2];
                covAverageImg(y, x)[2] = m_denoised(y, x)[1] * m_denoised(y, x)[2];
            }
    }


    Image3f smoothSqrAverageImg;
    smoothImage(smoothSqrAverageImg, sqrAverageImg, m_kernel);

    Image3f smoothCovAverageImg;
    if (fullCovar) 
        smoothImage(smoothCovAverageImg, covAverageImg, m_kernel);

    Image3f localNoiseVarImg;
    smoothImage(localNoiseVarImg, m_remainingNoiseVariance, m_kernel);

    RasterImage wallisFilteredImage;
    wallisFilteredImage.resize(width, height);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++) {

            Eigen::Vector3f localSignalVar = smoothSqrAverageImg(y, x) - smoothAverageImg(y, x).cwiseProduct(smoothAverageImg(y, x));

            int channels[3];

            if (fullCovar) {
                Eigen::Matrix3d covar;
                covar(0, 0) = localSignalVar[0];
                covar(1, 1) = localSignalVar[1];
                covar(2, 2) = localSignalVar[2];

                covar(0, 1) =
                covar(1, 0) = smoothCovAverageImg(y, x)[0] - smoothAverageImg(y, x)[0]*smoothAverageImg(y, x)[1];
                covar(0, 2) =
                covar(2, 0) = smoothCovAverageImg(y, x)[1] - smoothAverageImg(y, x)[0]*smoothAverageImg(y, x)[2];
                covar(1, 2) =
                covar(2, 1) = smoothCovAverageImg(y, x)[2] - smoothAverageImg(y, x)[1]*smoothAverageImg(y, x)[2];

                Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> es;
                es.compute(covar);

                const auto &V = es.eigenvectors();
                const auto &D = es.eigenvalues();

                const auto D_invHalf = Eigen::Vector3d(
                    ((D[0] < 1.0f)?0.0f:127.0f/3.0f/std::sqrt(D[0])),
                    ((D[1] < 1.0f)?0.0f:127.0f/3.0f/std::sqrt(D[1])),
                    ((D[2] < 1.0f)?0.0f:127.0f/3.0f/std::sqrt(D[2]))
                );


                const Eigen::Matrix3d whitener = V * Eigen::DiagonalMatrix<double, 3>(D_invHalf) * V.transpose();

                Eigen::Matrix3d whitenedNoiseCovar;
                for (unsigned i = 0; i < 3; i++)
                    for (unsigned j = 0; j < 3; j++) {
                        whitenedNoiseCovar(i, j) = 0.0f;
                        for (unsigned k = 0; k < 3; k++)
                            whitenedNoiseCovar(i, j) += whitener(i, k) * whitener(j, k) * localNoiseVarImg(y, x)[k]; // no covar in localNoiseVarImg
                    }

                Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> es2;
                es2.compute(whitenedNoiseCovar);

                const auto &Vn = es2.eigenvectors();
                const auto &Dn = es2.eigenvalues();

                const auto Dn_new = Eigen::Vector3d(
                    (std::sqrt(Dn[0]) > 10.0f/3.0f?10.0f/3.0f/std::sqrt(Dn[0]):1.0f),
                    (std::sqrt(Dn[1]) > 10.0f/3.0f?10.0f/3.0f/std::sqrt(Dn[1]):1.0f),
                    (std::sqrt(Dn[2]) > 10.0f/3.0f?10.0f/3.0f/std::sqrt(Dn[2]):1.0f)
                );


                const Eigen::Matrix3d M = Vn * Eigen::DiagonalMatrix<double, 3>(Dn_new) * Vn.transpose() * whitener;

                const Eigen::Vector3d v = M * Eigen::Vector3d((m_denoised(y, x) - smoothAverageImg(y, x)).cast<double>());
                for (unsigned c = 0; c < 3; c++)
                    channels[c] = std::max<int>(std::min<int>(127.5f + v[c], 255), 0);

            } else {
                for (unsigned c = 0; c < 3; c++) {
                    float maxSignalGain = 127.0f / std::max(1.0f, 3.0f * std::sqrt(localSignalVar[c]));
                    float maxNoiseGain = 10.0f / std::max(1.0f, 3.0f * std::sqrt(localNoiseVarImg(y, x)[c]));

                    float gain = std::min(maxSignalGain, maxNoiseGain);

                    channels[c] = std::max<int>(std::min<int>(127.5f + gain * (m_denoised(y, x)[c] - smoothAverageImg(y, x)[c]), 255), 0);
                }
            }

            wallisFilteredImage.getData()[x+y*width] =
                            (channels[0] << 0) |
                            (channels[1] << 8) |
                            (channels[2] << 16) |
                            (0xFF << 24);
        }

    return wallisFilteredImage;
}

}
