/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef DENOISER_H
#define DENOISER_H

#include "ImageHelpers.h"
#include <tools/RasterImage.h>

#include <string>
#include <vector>
#include <functional>

namespace whiteWalls {

/**
 * @brief Performs denoising and 8-bit packing of an image stack
 * @details Can be initialized with or without a linear FPN model.
 * Call Denoiser::denoise to compute the denoised image which can subsequently be accessed with Denoiser::getDenoisedImage.
 * After calling Denoiser::denoise, the denoised image can be packed into an 8-bit image using a modified wallis filter via Denoiser::wallisFilter.
 */
class Denoiser
{
    public:
        /// A callback function to report on the progress
        typedef std::function<void(const std::string &)> ProgressCallback;
        
        /**
         * @brief Initializes the denoiser
         * @param sigma Controls the width (in pixels) of the gaussian filter used in all blur operations. A value of 10.0 is usually appropriate.
         * @param fnpScale Scale factors of the optional linear fixed pattern noise model
         * @param fnpOffset Offsets of the optional linear fixed pattern noise model
         */
        Denoiser(float sigma, Image3f fpnScale = Image3f(), Image3f fpnOffset = Image3f());
        
        /**
         * @brief Runs the denoising procedure on a stack of images
         * @details The filenames should point to 16-bit images files containing the raw but demosaiced image data.
         * Tonemapped images are only experimentally supported through the tonemapping curve of an eos 400D.
         * @param filenames List of files that are to be stacked into one denoised image
         * @param linear Whether the image data is linear (directly from raw) or tonemapped (experimental jpg support)
         * @param progressCallback Optional callback to report on the progress
         */
        void denoise(const std::vector<std::string> &filenames, bool linear, ProgressCallback progressCallback = ProgressCallback());
        /// Returns the denoised image after Denoiser::denoise was called
        inline const Image3f &getDenoisedImage() const { return m_denoised; }

        /**
         * @brief Runs a modified wallis filter on the denoised image (after Denoiser::denoise was called) to pack the float data into an 8-bit image.
         * @details To estimate the residual noise and limit the amplification accordingly, the original images are needed a second time.
         * @param filenames The same list of files that Denoiser::denoise was called with
         * @param linear Same as in Denoiser::denoise
         * @param fullCovar Whether or not to normalize the full color covariance or each channel individually
         * @param progressCallback Optional callback to report on the progress
         * @returns Packed 8-bit image for use in the SfM/MVS pipeline
         */
        RasterImage wallisFilter(const std::vector<std::string> &filenames, bool linear, bool fullCovar = false, ProgressCallback progressCallback = ProgressCallback());
        /// Returns the estimated residual noise after Denoiser::wallisFilter was called
        inline const Image3f &getRemainingNoiseImage() const { return m_remainingNoiseVariance; }
    protected:
        Kernel m_kernel;
        Image3f m_fpnScale;
        Image3f m_fpnOffset;
        Image3f m_denoised;
        Image3f m_remainingNoiseVariance;
};

}

#endif // DENOISER_H
