/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "FPNCalibration.h"
#include "ImageHelpers.h"
#include "CameraResponseCurve.h"

#include <tools/TaskScheduler.h>
#include <tools/RasterImage.h>
#include <Magick++.h>
#include <boost/math/distributions.hpp>
#include <boost/format.hpp>

#include <ImfThreading.h>

#include <cstdint>
#include <mutex>

namespace whiteWalls {


FPNCalibration::BrightnessLevel::BrightnessLevel(std::vector<std::string> filenames, unsigned brightnessLevelIndex) :
        m_brightnessLevelIndex(brightnessLevelIndex),
        m_images(std::move(filenames))
{
    m_meanImage = (boost::format("%02d_mean.exr") % m_brightnessLevelIndex).str();
    m_varImage = (boost::format("%02d_var.exr") % m_brightnessLevelIndex).str();
    m_smoothedMeanImage = (boost::format("%02d_mean_smoothed.exr") % m_brightnessLevelIndex).str();
}

void FPNCalibration::BrightnessLevel::computeMeanImage()
{
    assert(!m_images.empty());
    unsigned width, height;
    RasterImage::getImageFileDimensions(m_images[0].c_str(), width, height);

    Image3f meanImg;
    meanImg.resize(height, width);
    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            meanImg(y, x).setZero();

    Image3f inputImg;
    for (unsigned i = 0; i < m_images.size(); i++) {
        readLinear16BitImage(m_images[i].c_str(), inputImg);

        meanImg += inputImg;
    }

    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            meanImg(y, x) /= m_images.size();

    writeExrImage(m_meanImage, meanImg, Imf::globalThreadCount());
}


void FPNCalibration::BrightnessLevel::smoothAvgImage(const Kernel &smoothingKernel)
{
    Image3f src, dst;

    readExrImage(m_meanImage, src, Imf::globalThreadCount());
    smoothImage(dst, src, smoothingKernel);
    writeExrImage(m_smoothedMeanImage, dst, Imf::globalThreadCount());
}

void FPNCalibration::BrightnessLevel::computeVariance(const Kernel &smoothingKernel)
{
    Image3f meanImg, smoothedMeanImg, varImg;

    readExrImage(m_meanImage, meanImg, Imf::globalThreadCount());
    readExrImage(m_smoothedMeanImage, smoothedMeanImg, Imf::globalThreadCount());

    varImg.resize(smoothedMeanImg.rows(), smoothedMeanImg.cols());
    for (unsigned y = 0; y < varImg.rows(); y++)
        for (unsigned x = 0; x < varImg.cols(); x++)
            varImg(y, x).setZero();

    Image3f img, smoothedImg;
    for (unsigned i = 0; i < m_images.size(); i++) {
        std::cout << m_brightnessLevelIndex << "  " << i << std::endl;
        readLinear16BitImage(m_images[i].c_str(), img);
        smoothImage(smoothedImg, img, smoothingKernel);

        for (unsigned y = 0; y < varImg.rows(); y++) {
            for (unsigned x = 0; x < varImg.cols(); x++) {

                Eigen::Array3f localBrightnessFactor;
                for (unsigned c = 0; c < 3; c++)
                    localBrightnessFactor[c] = smoothedMeanImg(y, x)[c] / std::max(1.0f, smoothedImg(y, x)[c]);

                Eigen::Array3f adjustedSample = img(y, x) * localBrightnessFactor;

                Eigen::Array3f error = adjustedSample - meanImg(y, x);

                varImg(y, x) += error * error;
            }
        }
    }

    for (unsigned y = 0; y < varImg.rows(); y++)
        for (unsigned x = 0; x < varImg.cols(); x++)
            varImg(y, x) /= m_images.size();

    writeExrImage(m_varImage, varImg, Imf::globalThreadCount());
}


FPNCalibration::FPNCalibration(std::vector<std::vector<std::string>> filenames)
{
    m_brightnessLevels.reserve(filenames.size());
    for (unsigned i = 0; i < filenames.size(); i++)
        m_brightnessLevels.push_back(BrightnessLevel(filenames[i], i));

    m_smoothingKernel = buildGaussianKernel(30.0f);
}

void FPNCalibration::computeMeanImage()
{
    // This step is pretty much disc IO bound so better not parallize it...
    for (unsigned i = 0; i < m_brightnessLevels.size(); i++) {
        std::cout << "Brightness level " << i << std::endl;
        m_brightnessLevels[i].computeMeanImage();
    }
}

void FPNCalibration::smoothAvgImage()
{
    for (unsigned i = 0; i < m_brightnessLevels.size(); i++)
        m_brightnessLevels[i].smoothAvgImage(m_smoothingKernel);
}

void FPNCalibration::computeVariance()
{
    for (unsigned i = 0; i < m_brightnessLevels.size(); i++)
        m_brightnessLevels[i].computeVariance(m_smoothingKernel);
}


struct LinearFit {
    Eigen::Matrix2d m;
    Eigen::Vector2d inhom;

    LinearFit() {
        m.setZero();
        inhom.setZero();
    }
    void add(float x, float y, float weight) {
        m(0, 0) += x*x * weight;
        m(0, 1) += x * weight;
        m(1, 0) += x * weight;
        m(1, 1) += 1.0f * weight;

        inhom[0] += x * y * weight;
        inhom[1] += y * weight;
    }
    void solve(float &a, float &b) {
        Eigen::Vector2d solution = m.fullPivLu().solve(inhom);
        a = solution[0];
        b = solution[1];
    }
};


void FPNCalibration::computeLinearModel(Image3f &scales, Image3f &offsets)
{
    std::vector<Image3f> means, smoothedMeans, vars;
    means.resize(m_brightnessLevels.size());
    smoothedMeans.resize(m_brightnessLevels.size());
    vars.resize(m_brightnessLevels.size());

    for (unsigned i = 0; i < m_brightnessLevels.size(); i++) {
        readExrImage(m_brightnessLevels[i].getMeanImage(), means[i], Imf::globalThreadCount());
        readExrImage(m_brightnessLevels[i].getSmoothedMeanImage(), smoothedMeans[i], Imf::globalThreadCount());
        readExrImage(m_brightnessLevels[i].getVarImage(), vars[i], Imf::globalThreadCount());
    }


    offsets.resize(means[0].rows(), means[0].cols());
    scales.resize(means[0].rows(), means[0].cols());

    for (unsigned y = 0; y < offsets.rows(); y++)
        for (unsigned x = 0; x < offsets.cols(); x++) {

            for (unsigned channel = 0; channel < 3; channel++) {
                LinearFit meanLinFit;

                for (unsigned j = 0; j < m_brightnessLevels.size(); j++) {
                    float pred = smoothedMeans[j](y, x)[channel];
                    float resp = means[j](y, x)[channel];

                    float vagueness = std::max(1.0f, vars[j](y, x)[channel]) / m_brightnessLevels[j].getImageCount();

                    if (pred < 60000) // skip saturated/clamped channels
                        meanLinFit.add(pred, resp, 1.0f / vagueness);
                }

                meanLinFit.solve(scales(y, x)[channel], offsets(y, x)[channel]);
            }
        }
}

void FPNCalibration::produceDebugImgOfScale(Image3f &scales, const std::string &filename)
{
    std::vector<float> sortedScales;
    for (unsigned y = 0; y < scales.rows(); y++)
        for (unsigned x = 0; x < scales.cols(); x++)
            for (unsigned channel = 0; channel < 3; channel++)
                sortedScales.push_back(scales(y, x)[channel]);

    std::sort(sortedScales.begin(), sortedScales.end());

    float scaleRangeMin = sortedScales[sortedScales.size() / 10];
    float scaleRangeMax = sortedScales[sortedScales.size() * 9 / 10];

    std::cout << "Scale between: " << scaleRangeMin << " and " << scaleRangeMax << " (10% and 90% quantiles)" << std::endl;

    const unsigned width = scales.cols();
    const unsigned height = scales.rows();

    RasterImage scaleImage;
    scaleImage.resize(width, height);

    for (unsigned y = 0; y < height; y++) {
        for (unsigned x = 0; x < width; x++) {
            int ir = (scales(y, x)[0] - 1.0f) / std::max(std::abs(1.0f - scaleRangeMax), std::abs(1.0f - scaleRangeMin)) * 128 + 128;
            int ig = (scales(y, x)[1] - 1.0f) / std::max(std::abs(1.0f - scaleRangeMax), std::abs(1.0f - scaleRangeMin)) * 128 + 128;
            int ib = (scales(y, x)[2] - 1.0f) / std::max(std::abs(1.0f - scaleRangeMax), std::abs(1.0f - scaleRangeMin)) * 128 + 128;

            ir = std::min<int>(std::max<int>(ir, 0), 255);
            ig = std::min<int>(std::max<int>(ig, 0), 255);
            ib = std::min<int>(std::max<int>(ib, 0), 255);

            scaleImage.getData()[x+y*width] =
                        (ir << 0) |
                        (ig << 8) |
                        (ib << 16) |
                        (0xFF << 24);
        }
    }
    scaleImage.writeToFile(filename.c_str());
}

void FPNCalibration::produceDebugImgOfOffset(Image3f &offsets, const std::string &filename)
{
    std::vector<float> sortedOffsets;
    for (unsigned y = 0; y < offsets.rows(); y++)
        for (unsigned x = 0; x < offsets.cols(); x++)
            for (unsigned channel = 0; channel < 3; channel++)
                sortedOffsets.push_back(offsets(y, x)[channel]);


    std::sort(sortedOffsets.begin(), sortedOffsets.end());

    float offsetRangeMin = sortedOffsets[sortedOffsets.size() / 10];
    float offsetRangeMax = sortedOffsets[sortedOffsets.size() * 9 / 10];

    std::cout << "Offset between: " << offsetRangeMin<<"/65535" << " and " << offsetRangeMax<<"/65535" << " (10% and 90% quantiles)" << std::endl;

    const unsigned width = offsets.cols();
    const unsigned height = offsets.rows();

    RasterImage offsetImage;
    offsetImage.resize(width, height);
    RasterImage scaleImage;
    scaleImage.resize(width, height);
    for (unsigned y = 0; y < height; y++) {
        for (unsigned x = 0; x < width; x++) {
            int ir = offsets(y, x)[0] / std::max(std::abs(offsetRangeMax), std::abs(offsetRangeMin)) * 128 + 128;
            int ig = offsets(y, x)[1] / std::max(std::abs(offsetRangeMax), std::abs(offsetRangeMin)) * 128 + 128;
            int ib = offsets(y, x)[2] / std::max(std::abs(offsetRangeMax), std::abs(offsetRangeMin)) * 128 + 128;

            ir = std::min<int>(std::max<int>(ir, 0), 255);
            ig = std::min<int>(std::max<int>(ig, 0), 255);
            ib = std::min<int>(std::max<int>(ib, 0), 255);

            offsetImage.getData()[x+y*width] =
                        (ir << 0) |
                        (ig << 8) |
                        (ib << 16) |
                        (0xFF << 24);
        }
    }
    offsetImage.writeToFile(filename.c_str());
}


void FPNCalibration::produceHistogram(Image3f &img, float min, float max, unsigned numBars, std::ostream &stream)
{
    std::vector<float> allOfEm;
    for (unsigned y = 0; y < img.rows(); y++)
        for (unsigned x = 0; x < img.cols(); x++)
            for (unsigned channel = 0; channel < 3; channel++)
                allOfEm.push_back(img(y, x)[channel]);


    std::vector<unsigned> bars;
    bars.resize(numBars);

    float scale = numBars / (max - min);
    float offset = -min * scale;

    for (float f : allOfEm) {
        int index = std::floor(f * scale + offset);
        if ((index < 0) || (index >= (int)numBars)) continue;

        bars[index]++;
    }

    stream << "x = [";
    for (unsigned i = 0; i < numBars; i++)
        stream << (i+0.5f - offset) / scale << ' ';

    stream << "];" << std::endl;

    stream << "y = [";
    for (unsigned i = 0; i < numBars; i++)
        stream << bars[i] / (float) allOfEm.size() << ' ';

    stream << "];" << std::endl;

    stream << "bar(x, y, \"hist\");" << std::endl;
}

}