/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef FPNCALIBRATION_H
#define FPNCALIBRATION_H

#include "ImageHelpers.h"

#include <vector>
#include <string>
#include <ostream>

namespace whiteWalls {

/**
 * @brief FPN calibration code
 * @details This class performs a sequence of steps to compute the linear FPN model from a set of calibration images.
 * The method creates files for the intermediate results so that a rerun does not require the repetition of all steps.
 * This facilitates experimentation and debuggin a lot.
 * After construction, the following methods have to be called in this order:
 *    * FPNCalibration::computeMeanImage();
 *    * FPNCalibration::smoothAvgImage();
 *    * FPNCalibration::computeVariance();
 *    * FPNCalibration::computeLinearModel(Image3f &scales, Image3f &offsets);
 *
 * The last call produces the estimated scales and offsets that can be used in the @ref Denoiser "image denoiser".
 */
class FPNCalibration
{
    public:
        /**
         * @brief Setup the calibration with a dataset.
         * @param filenames Filenames of the 16-bit (possibly tiff) images for calibration.
         * @details The outer vector's size is equal to the number of different shutter speeds for which images were captured.
         * For each shutter speed, the outer vector contains an inner vector of filenames, one filename for each image with that shutter speed.
         */
        FPNCalibration(std::vector<std::vector<std::string>> filenames);

        /// Computes the mean images for each brightness level
        void computeMeanImage();
        /// Computes the estimated "true" intensities by bluring the mean images
        void smoothAvgImage();
        /// Estimates the remaining variance in the mean images of each brightness level so that weights can be derived from them.
        void computeVariance();

        /// Performs weighted least squares fits of linear functions between estimated intensities and outputs for each pixel and channel.
        void computeLinearModel(Image3f &scales, Image3f &offsets);

        /// Produces a debug image of the scales to show the patterns
        static void produceDebugImgOfScale(Image3f &scales, const std::string &filename);
        /// Produces a debug image of the offsets to show the patterns
        static void produceDebugImgOfOffset(Image3f &offsets, const std::string &filename);

        /// Produces matlab/octave code that plots a histogra, of the values in the image
        static void produceHistogram(Image3f &img, float min, float max, unsigned numBars, std::ostream &stream);
    protected:
        Kernel m_smoothingKernel;

        class BrightnessLevel
        {
            public:
                BrightnessLevel(std::vector<std::string> filenames, unsigned brightnessLevelIndex);

                void computeMeanImage();
                void smoothAvgImage(const Kernel &smoothingKernel);
                void computeVariance(const Kernel &smoothingKernel);

                inline const std::string &getMeanImage() const { return m_meanImage; }
                inline const std::string &getVarImage() const { return m_varImage; }
                inline const std::string &getSmoothedMeanImage() const { return m_smoothedMeanImage; }

                inline unsigned getImageCount() const { return m_images.size(); }
            protected:
                unsigned m_brightnessLevelIndex;
                std::vector<std::string> m_images;
                std::string m_meanImage;
                std::string m_varImage;
                std::string m_smoothedMeanImage;
        };

        std::vector<BrightnessLevel> m_brightnessLevels;
};

}

#endif // FPNCALIBRATION_H
