/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "ImageHelpers.h"
#include <tools/TaskScheduler.h>

#include <ImfInputFile.h>
#include <ImfArray.h>
#include <ImathBox.h>
#include <ImfOutputFile.h>
#include <ImfChannelList.h>
#include <ImfThreading.h>

#include <Magick++.h>

#include <stdexcept>
#include <cmath>

namespace whiteWalls {


void smoothImage(Image3f &dst, const Image3f &src, const Kernel &kernel)
{
    Image3f tmp;
    tmp.resize(src.rows(), src.cols());
    dst.resize(src.rows(), src.cols());
    {
        TaskGroup group;
        for (unsigned y = 0; y < src.rows(); y++) {
            group.add([y, &kernel, &src, &tmp]{
                for (unsigned x = 0; x < src.cols(); x++) {
                    Eigen::Array3f sum;
                    sum.setZero();

                    for (int i = -(int)kernel.size()/2; i <= (int)kernel.size()/2; i++) {
                        int y_ = std::min<int>(std::max<int>(((int)y)+i, 0), src.rows()-1);

                        sum += kernel[i+(int)kernel.size()/2] * src(y_, x);
                    }
                    tmp(y, x) = sum;
                }
            }, TaskScheduler::get());
        }
        TaskScheduler::get().waitFor(&group);
    }

    {
        TaskGroup group;
        for (unsigned y = 0; y < tmp.rows(); y++) {
            group.add([y, &kernel, &tmp, &dst]{
                for (unsigned x = 0; x < tmp.cols(); x++) {
                    Eigen::Array3f sum;
                    sum.setZero();

                    for (int i = -(int)kernel.size()/2; i <= (int)kernel.size()/2; i++) {
                        int x_ = std::min<int>(std::max<int>((int)x+i, 0), tmp.cols()-1);

                        sum += kernel[i+(int)kernel.size()/2] * tmp(y, x_);
                    }
                    dst(y, x) = sum;
                }
            }, TaskScheduler::get());
        }
        TaskScheduler::get().waitFor(&group);
    }
}

Kernel buildGaussianKernel(float sigma)
{
    Kernel kernel;
    int w = sigma*3;
    kernel.resize(2*w+1);
    for (unsigned i = 0; i < kernel.size(); i++) {
        int j = (int)i - w;
        kernel[i] = 1.0f / (sigma * std::sqrt(2.0f * (float)M_PI)) * std::exp(-(float)(j*j)/(2.0f * sigma * sigma));
    }
    normalizeFilterKernel(kernel);
    return kernel;
}

void normalizeFilterKernel(Kernel &kernel)
{
    float sum = 0.0f;
    for (unsigned i = 0; i < kernel.size(); i++)
        sum += kernel[i];
    float fac = 1.0f / sum;
    for (unsigned i = 0; i < kernel.size(); i++)
        kernel[i] *= fac;
}


void readExrImage(const std::string &filename, Image3f &image, unsigned threads)
{
    Imf::InputFile inputFile(filename.c_str(), threads);

    Imath::Box2i dw = inputFile.header().dataWindow();
    unsigned width = dw.max.x - dw.min.x + 1;
    unsigned height = dw.max.y - dw.min.y + 1;
    const int dx = dw.min.x;
    const int dy = dw.min.y;

    if ((dx != 0) ||(dy != 0))
        throw std::runtime_error("dx != 0 or dy != 0 not supported!");

    image.resize(height, width);

    Imf::FrameBuffer frameBuffer;
    frameBuffer.insert("R", Imf::Slice(Imf::FLOAT, ((char *) &image(0, 0)[0]), sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f)*width));
    frameBuffer.insert("G", Imf::Slice(Imf::FLOAT, ((char *) &image(0, 0)[1]), sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f)*width));
    frameBuffer.insert("B", Imf::Slice(Imf::FLOAT, ((char *) &image(0, 0)[2]), sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f)*width));

    inputFile.setFrameBuffer (frameBuffer);
    inputFile.readPixels(0, height-1);
}

void writeExrImage(const std::string &filename, const Image3f &image, unsigned threads)
{
    Imf::Header outputHeader(image.cols(), image.rows());
    outputHeader.channels().insert("R", Imf::Channel(Imf::FLOAT));
    outputHeader.channels().insert("G", Imf::Channel(Imf::FLOAT));
    outputHeader.channels().insert("B", Imf::Channel(Imf::FLOAT));

    Imf::FrameBuffer outputframeBuffer;
    outputframeBuffer.insert("R", Imf::Slice(Imf::FLOAT, (char *) &image(0, 0)[0], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * image.cols()));
    outputframeBuffer.insert("G", Imf::Slice(Imf::FLOAT, (char *) &image(0, 0)[1], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * image.cols()));
    outputframeBuffer.insert("B", Imf::Slice(Imf::FLOAT, (char *) &image(0, 0)[2], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * image.cols()));

    Imf::OutputFile outputFile(filename.c_str(), outputHeader, threads);
    outputFile.setFrameBuffer(outputframeBuffer);
    outputFile.writePixels(image.rows());
}

template<class ColorToIrradiance>
void read16BitImage(const std::string &filename, Image3f &img, ColorToIrradiance colorToIrradiance)
{
    Magick::Image imageFile(filename.c_str());

    unsigned width = imageFile.columns();
    unsigned height = imageFile.rows();

    img.resize(height, width);

    std::vector<std::array<unsigned short, 4>> rgba16;
    rgba16.resize(width*height);
    imageFile.write(0, 0, imageFile.columns(), imageFile.rows(), "RGBA", Magick::ShortPixel, rgba16.data());

    for (unsigned y = 0; y < height; y++)
        for (unsigned x = 0; x < width; x++)
            img(y, x) = colorToIrradiance(rgba16[x+y*width]);
}


void readLinear16BitImage(const std::string &filename, Image3f &img)
{
    read16BitImage(filename, img, [](const std::array<unsigned short, 4> &color)->Eigen::Array3f{
        return Eigen::Array3f(color[0], color[1], color[2]);
    });
}

void readTonemappedImage(const std::string &filename, Image3f &img, 
                             const synthBench::math::CameraResponseCurve<256> &red,
                             const synthBench::math::CameraResponseCurve<256> &green,
                             const synthBench::math::CameraResponseCurve<256> &blue)
{
    read16BitImage(filename, img, [&red, &green, &blue](const std::array<unsigned short, 4> &color)->Eigen::Array3f{
        float irradianceR = red.ldrToHdr(std::min(color[0] >> 8, 255)) * 65535;
        float irradianceG = green.ldrToHdr(std::min(color[1] >> 8, 255)) * 65535;
        float irradianceB = blue.ldrToHdr(std::min(color[2] >> 8, 255)) * 65535;
        
        return Eigen::Array3f(irradianceR, irradianceG, irradianceB);        
    });
}

void removeLinearFPN(Image3f &img,
               const Image3f &fpnScale,
               const Image3f &fpnOffset)
{
    if ((img.rows() != fpnScale.rows()) ||
        (img.cols() != fpnScale.cols()) ||
        (img.rows() != fpnOffset.rows()) ||
        (img.cols() != fpnOffset.cols()))
        throw std::runtime_error("Image dimensions mismatch!");
        
    for (unsigned y = 0; y < img.rows(); y++)
        for (unsigned x = 0; x < img.cols(); x++)
            for (unsigned c = 0; c < 3; c++)
                img(y, x)[c] = (img(y, x)[c] / std::max(0.000001f, fpnScale(y, x)[c])) - fpnOffset(y, x)[c]; // It's a good thing we don't care about performance...
}

}