/*
 * White Walls - Fuse multiple photographs to reduce noise and pack into 8-Bit image for SfM/MVS
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


#ifndef IMAGEHELPERS_H
#define IMAGEHELPERS_H

#include "CameraResponseCurve.h"

#include <Eigen/Dense>
#include <vector>

/// All functions and classes specific to the White Walls paper
namespace whiteWalls {

/// A simple RGB32F image
typedef Eigen::Array<Eigen::Array3f, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor> Image3f;
/// 1D smoothing kernel
typedef std::vector<float> Kernel;

/// Smooth an image by convolving with the kernel vertically and then horizontally
void smoothImage(Image3f &dst, const Image3f &src, const Kernel &kernel);

/// Builds a kernel for gaussian blur
Kernel buildGaussianKernel(float sigma);
/// Normalizes a kernel to sum to one
void normalizeFilterKernel(Kernel &kernel);

/// Reads a RGB32F image from an OpenEXR .exr file
void readExrImage(const std::string &filename, Image3f &image, unsigned threads = 1);
/// Writes a RGB32F image to an OpenEXR .exr file
void writeExrImage(const std::string &filename, const Image3f &image, unsigned threads = 1);

/// Reads an ordinary (png, tiff) image with 16-bit precision
void readLinear16BitImage(const std::string &filename, Image3f &img);

/// @brief Reads an 8-bit image and applies pre-calibrated camera tonemapping curves to transform the data back into a linear color space
/// @details We experimented with this but did not do extensive testing of linear vs. sRGB spaces because RAW data worked just so much better.
void readTonemappedImage(const std::string &filename, Image3f &img, 
                             const synthBench::math::CameraResponseCurve<256> &red,
                             const synthBench::math::CameraResponseCurve<256> &green,
                             const synthBench::math::CameraResponseCurve<256> &blue);

/// Removes fixed pattern noise (FPN) from an image using the linear FPN model
void removeLinearFPN(Image3f &img,
               const Image3f &fpnScale,
               const Image3f &fpnOffset);

}

#endif // IMAGEHELPERS_H
